@@ -10,7 +10,7 @@
 /*	CLEANUP_STATE *state;
 /*	int	type;
 /*	const char *buf;
-/*	int	len;
+/*	ssize_t	len;
 /* DESCRIPTION
 /*	This module processes envelope records and writes the result
 /*	to the queue file.  It validates the message structure, rewrites
@@ -44,10 +44,7 @@
 #include <sys_defs.h>
 #include <string.h>
 #include <stdlib.h>
-
-#ifdef STRCASECMP_IN_STRINGS_H
-#include <strings.h>
-#endif
+#include <ctype.h>
 
 /* Utility library. */
 
@@ -67,6 +64,8 @@
 #include <mail_params.h>
 #include <verp_sender.h>
 #include <mail_proto.h>
+#include <dsn_mask.h>
+#include <rec_attr_map.h>
 
 /* Application-specific. */
 
@@ -75,12 +74,12 @@
 #define STR	vstring_str
 #define STREQ(x,y) (strcmp((x), (y)) == 0)
 
-static void cleanup_envelope_process(CLEANUP_STATE *, int, const char *, int);
+static void cleanup_envelope_process(CLEANUP_STATE *, int, const char *, ssize_t);
 
 /* cleanup_envelope - initialize message envelope */
 
 void    cleanup_envelope(CLEANUP_STATE *state, int type,
-			         const char *str, int len)
+			         const char *str, ssize_t len)
 {
 
     /*
@@ -93,7 +92,7 @@ void    cleanup_envelope(CLEANUP_STATE *state, int type,
 		       (REC_TYPE_SIZE_CAST1) 0,	/* content size */
 		       (REC_TYPE_SIZE_CAST2) 0,	/* content offset */
 		       (REC_TYPE_SIZE_CAST3) 0,	/* recipient count */
-		       (REC_TYPE_SIZE_CAST4) 0);/* qmgr options */
+		       (REC_TYPE_SIZE_CAST4) 0);	/* qmgr options */
 
     /*
      * Pass control to the actual envelope processing routine.
@@ -105,19 +104,29 @@ void    cleanup_envelope(CLEANUP_STATE *state, int type,
 /* cleanup_envelope_process - process one envelope record */
 
 static void cleanup_envelope_process(CLEANUP_STATE *state, int type,
-				             const char *buf, int len)
+				             const char *buf, ssize_t len)
 {
+    const char *myname = "cleanup_envelope_process";
     char   *attr_name;
     char   *attr_value;
     const char *error_text;
     int     extra_opts;
+    int     junk;
+    int     mapped_type = type;
+    const char *mapped_buf = buf;
+    int     milter_count;
+
+#ifdef DELAY_ACTION
+    int     defer_delay;
+
+#endif
 
     if (msg_verbose)
-	msg_info("initial envelope %c %.*s", type, len, buf);
+	msg_info("initial envelope %c %.*s", type, (int) len, buf);
 
     if (type == REC_TYPE_FLGS) {
 	/* Not part of queue file format. */
-	extra_opts = atol(buf);
+	extra_opts = atoi(buf);
 	if (extra_opts & ~CLEANUP_FLAG_MASK_EXTRA)
 	    msg_warn("%s: ignoring bad extra flags: 0x%x",
 		     state->queue_id, extra_opts);
@@ -125,6 +134,54 @@ static void cleanup_envelope_process(CLEANUP_STATE *state, int type,
 	    state->flags |= extra_opts;
 	return;
     }
+#ifdef DELAY_ACTION
+    if (type == REC_TYPE_DELAY) {
+	/* Not part of queue file format. */
+	defer_delay = atoi(buf);
+	if (defer_delay <= 0)
+	    msg_warn("%s: ignoring bad delay time: %s", state->queue_id, buf);
+	else
+	    state->defer_delay = defer_delay;
+	return;
+    }
+#endif
+    if (type == REC_TYPE_MILT_COUNT) {
+	/* Not part of queue file format. */
+	if ((milter_count = atoi(buf)) >= 0)
+	    cleanup_milter_receive(state, milter_count);
+	return;
+    }
+
+    /*
+     * Map DSN attribute name to pseudo record type so that we don't have to
+     * pollute the queue file with records that are incompatible with past
+     * Postfix versions. Preferably, people should be able to back out from
+     * an upgrade without losing mail.
+     */
+    if (type == REC_TYPE_ATTR) {
+	vstring_strcpy(state->attr_buf, buf);
+	error_text = split_nameval(STR(state->attr_buf), &attr_name, &attr_value);
+	if (error_text != 0) {
+	    msg_warn("%s: message rejected: malformed attribute: %s: %.100s",
+		     state->queue_id, error_text, buf);
+	    state->errs |= CLEANUP_STAT_BAD;
+	    return;
+	}
+	/* Zero-length values are place holders for unavailable values. */
+	if (*attr_value == 0) {
+	    msg_warn("%s: spurious null attribute value for \"%s\" -- ignored",
+		     state->queue_id, attr_name);
+	    return;
+	}
+	if ((junk = rec_attr_map(attr_name)) != 0) {
+	    mapped_buf = attr_value;
+	    mapped_type = junk;
+	}
+    }
+
+    /*
+     * Sanity check.
+     */
     if (strchr(REC_TYPE_ENVELOPE, type) == 0) {
 	msg_warn("%s: message rejected: unexpected record type %d in envelope",
 		 state->queue_id, type);
@@ -164,16 +221,23 @@ static void cleanup_envelope_process(CLEANUP_STATE *state, int type,
 	    state->errs |= CLEANUP_STAT_BAD;
 	    return;
 	}
-	if (state->time == 0) {
+	if (state->arrival_time.tv_sec == 0) {
 	    msg_warn("%s: message rejected: missing time envelope record",
 		     state->queue_id);
 	    state->errs |= CLEANUP_STAT_BAD;
 	    return;
 	}
+
+	/*
+	 * XXX This works by accident, because the sender is recorded at the
+	 * beginning of the envelope segment.
+	 */
 	if ((state->flags & CLEANUP_FLAG_WARN_SEEN) == 0
+	    && state->sender && *state->sender
 	    && var_delay_warn_time > 0) {
 	    cleanup_out_format(state, REC_TYPE_WARN, REC_TYPE_WARN_FORMAT,
-			       (long) state->time + var_delay_warn_time);
+			       REC_TYPE_WARN_ARG(state->arrival_time.tv_sec
+						 + var_delay_warn_time));
 	}
 	state->flags |= CLEANUP_FLAG_INRCPT;
     }
@@ -191,31 +255,77 @@ static void cleanup_envelope_process(CLEANUP_STATE *state, int type,
 	if (state->orig_rcpt == 0)
 	    state->orig_rcpt = mystrdup(buf);
 	cleanup_addr_recipient(state, buf);
+	if (cleanup_milters != 0
+	    && state->milters == 0
+	    && CLEANUP_MILTER_OK(state))
+	    cleanup_milter_emul_rcpt(state, cleanup_milters, buf);
 	myfree(state->orig_rcpt);
 	state->orig_rcpt = 0;
+	if (state->dsn_orcpt != 0) {
+	    myfree(state->dsn_orcpt);
+	    state->dsn_orcpt = 0;
+	}
+	state->dsn_notify = 0;
 	return;
     }
-    if (type == REC_TYPE_DONE) {
+    if (type == REC_TYPE_DONE || type == REC_TYPE_DRCP) {
 	if (state->orig_rcpt != 0) {
 	    myfree(state->orig_rcpt);
 	    state->orig_rcpt = 0;
 	}
+	if (state->dsn_orcpt != 0) {
+	    myfree(state->dsn_orcpt);
+	    state->dsn_orcpt = 0;
+	}
+	state->dsn_notify = 0;
 	return;
     }
-    if (state->orig_rcpt != 0) {
-	/* REC_TYPE_ORCP must be followed by REC_TYPE_RCPT or REC_TYPE DONE. */
-	msg_warn("%s: ignoring out-of-order original recipient record <%.200s>",
-		 state->queue_id, state->orig_rcpt);
-	myfree(state->orig_rcpt);
-	state->orig_rcpt = 0;
+    if (mapped_type == REC_TYPE_DSN_ORCPT) {
+	if (state->dsn_orcpt) {
+	    msg_warn("%s: ignoring out-of-order DSN original recipient record <%.200s>",
+		     state->queue_id, state->dsn_orcpt);
+	    myfree(state->dsn_orcpt);
+	}
+	state->dsn_orcpt = mystrdup(mapped_buf);
+	return;
+    }
+    if (mapped_type == REC_TYPE_DSN_NOTIFY) {
+	if (state->dsn_notify) {
+	    msg_warn("%s: ignoring out-of-order DSN notify record <%d>",
+		     state->queue_id, state->dsn_notify);
+	    state->dsn_notify = 0;
+	}
+	if (!alldig(mapped_buf) || (junk = atoi(mapped_buf)) == 0
+	    || DSN_NOTIFY_OK(junk) == 0)
+	    msg_warn("%s: ignoring malformed DSN notify record <%.200s>",
+		     state->queue_id, buf);
+	else
+	    state->qmgr_opts |=
+		QMGR_READ_FLAG_FROM_DSN(state->dsn_notify = junk);
+	return;
     }
     if (type == REC_TYPE_ORCP) {
+	if (state->orig_rcpt != 0) {
+	    msg_warn("%s: ignoring out-of-order original recipient record <%.200s>",
+		     state->queue_id, state->orig_rcpt);
+	    myfree(state->orig_rcpt);
+	}
 	state->orig_rcpt = mystrdup(buf);
 	return;
     }
     if (type == REC_TYPE_MESG) {
 	state->action = cleanup_message;
-	state->flags &= ~CLEANUP_FLAG_INRCPT;
+	if (state->flags & CLEANUP_FLAG_INRCPT) {
+	    if (state->milters || cleanup_milters) {
+		/* Make room to append recipient. */
+		if ((state->append_rcpt_pt_offset = vstream_ftell(state->dst)) < 0)
+		    msg_fatal("%s: vstream_ftell %s: %m:", myname, cleanup_path);
+		cleanup_out_format(state, REC_TYPE_PTR, REC_TYPE_PTR_FORMAT, 0L);
+		if ((state->append_rcpt_pt_target = vstream_ftell(state->dst)) < 0)
+		    msg_fatal("%s: vstream_ftell %s: %m:", myname, cleanup_path);
+	    }
+	    state->flags &= ~CLEANUP_FLAG_INRCPT;
+	}
 	return;
     }
 
@@ -228,12 +338,18 @@ static void cleanup_envelope_process(CLEANUP_STATE *state, int type,
     if (type == REC_TYPE_SIZE)
 	/* Use our own SIZE record instead. */
 	return;
+    if (mapped_type == REC_TYPE_CTIME)
+	/* Use our own expiration time base record instead. */
+	return;
     if (type == REC_TYPE_TIME) {
 	/* First instance wins. */
-	if (state->time == 0) {
-	    state->time = atol(buf);
+	if (state->arrival_time.tv_sec == 0) {
+	    REC_TYPE_TIME_SCAN(buf, state->arrival_time);
 	    cleanup_out(state, type, buf, len);
 	}
+	/* Generate our own expiration time base record. */
+	cleanup_out_format(state, REC_TYPE_ATTR, "%s=%ld",
+			   MAIL_ATTR_CREATE_TIME, (long) time((time_t *) 0));
 	return;
     }
     if (type == REC_TYPE_FULL) {
@@ -253,6 +369,47 @@ static void cleanup_envelope_process(CLEANUP_STATE *state, int type,
 	    return;
 	}
 	cleanup_addr_sender(state, buf);
+	if (cleanup_milters != 0
+	    && state->milters == 0
+	    && CLEANUP_MILTER_OK(state))
+	    cleanup_milter_emul_mail(state, cleanup_milters, buf);
+	return;
+    }
+    if (mapped_type == REC_TYPE_DSN_ENVID) {
+	/* Allow only one instance. */
+	if (state->dsn_envid != 0) {
+	    msg_warn("%s: message rejected: multiple DSN envelope ID records",
+		     state->queue_id);
+	    state->errs |= CLEANUP_STAT_BAD;
+	    return;
+	}
+	if (!allprint(mapped_buf)) {
+	    msg_warn("%s: message rejected: bad DSN envelope ID record",
+		     state->queue_id);
+	    state->errs |= CLEANUP_STAT_BAD;
+	    return;
+	}
+	state->dsn_envid = mystrdup(mapped_buf);
+	cleanup_out(state, type, buf, len);
+	return;
+    }
+    if (mapped_type == REC_TYPE_DSN_RET) {
+	/* Allow only one instance. */
+	if (state->dsn_ret != 0) {
+	    msg_warn("%s: message rejected: multiple DSN RET records",
+		     state->queue_id);
+	    state->errs |= CLEANUP_STAT_BAD;
+	    return;
+	}
+	if (!alldig(mapped_buf) || (junk = atoi(mapped_buf)) == 0
+	    || DSN_RET_OK(junk) == 0) {
+	    msg_warn("%s: message rejected: bad DSN RET record <%.200s>",
+		     state->queue_id, buf);
+	    state->errs |= CLEANUP_STAT_BAD;
+	    return;
+	}
+	state->dsn_ret = junk;
+	cleanup_out(state, type, buf, len);
 	return;
     }
     if (type == REC_TYPE_WARN) {
@@ -263,39 +420,44 @@ static void cleanup_envelope_process(CLEANUP_STATE *state, int type,
 	}
 	return;
     }
+    /* XXX Needed for cleanup_bounce(); sanity check usage. */
+    if (type == REC_TYPE_VERP) {
+	if (state->verp_delims == 0) {
+	    if (state->sender == 0 || state->sender[0] == 0) {
+		msg_warn("%s: ignoring VERP request for null sender",
+			 state->queue_id);
+	    } else if (verp_delims_verify(buf) != 0) {
+		msg_warn("%s: ignoring bad VERP request: \"%.100s\"",
+			 state->queue_id, buf);
+	    } else {
+		state->verp_delims = mystrdup(buf);
+		cleanup_out(state, type, buf, len);
+	    }
+	}
+	return;
+    }
     if (type == REC_TYPE_ATTR) {
-	char   *sbuf;
-
 	if (state->attr->used >= var_qattr_count_limit) {
 	    msg_warn("%s: message rejected: attribute count exceeds limit %d",
 		     state->queue_id, var_qattr_count_limit);
 	    state->errs |= CLEANUP_STAT_BAD;
 	    return;
 	}
-	sbuf = mystrdup(buf);
-	if ((error_text = split_nameval(sbuf, &attr_name, &attr_value)) != 0) {
-	    msg_warn("%s: message rejected: malformed attribute: %s: %.100s",
-		     state->queue_id, error_text, buf);
-	    state->errs |= CLEANUP_STAT_BAD;
-	    myfree(sbuf);
-	    return;
-	}
 	if (strcmp(attr_name, MAIL_ATTR_RWR_CONTEXT) == 0) {
 	    /* Choose header rewriting context. See also cleanup_addr.c. */
 	    if (STREQ(attr_value, MAIL_ATTR_RWR_LOCAL)) {
 		state->hdr_rewrite_context = MAIL_ATTR_RWR_LOCAL;
 	    } else if (STREQ(attr_value, MAIL_ATTR_RWR_REMOTE)) {
-		state->hdr_rewrite_context = 
-		    (*var_remote_rwr_domain ?  MAIL_ATTR_RWR_REMOTE : 0);
+		state->hdr_rewrite_context =
+		    (*var_remote_rwr_domain ? MAIL_ATTR_RWR_REMOTE : 0);
 	    } else {
-                msg_warn("%s: message rejected: bad rewriting context: %.100s",
-                         state->queue_id, attr_value);
-                state->errs |= CLEANUP_STAT_BAD;
-                return;
+		msg_warn("%s: message rejected: bad rewriting context: %.100s",
+			 state->queue_id, attr_value);
+		state->errs |= CLEANUP_STAT_BAD;
+		return;
 	    }
 	}
 	nvtable_update(state->attr, attr_name, attr_value);
-	myfree(sbuf);
 	cleanup_out(state, type, buf, len);
 	return;
     } else {
